from pathlib import Path
from typing import Dict, List

from pydantic import BaseModel
from strictyaml import YAML, load

import regression_model

# Project Directories
PACKAGE_ROOT = Path(regression_model.__file__).resolve().parent
ROOT = PACKAGE_ROOT.parent
CONFIG_FILE_PATH = PACKAGE_ROOT / "config.yml"
DATASET_DIR = PACKAGE_ROOT / "datasets"
TRAINED_MODEL_DIR = PACKAGE_ROOT / "trained_model"


class AppConfig(BaseModel):
    """
    Application-level config
    """

    package_name: str
    training_data_file: str
    test_data_file: str
    pipeline_save_file: str


class ModelConfig(BaseModel):
    """
    All configuration relevant to model
    training and feature engineering
    """

    target: str
    features: List[str]
    dropped_feature: str
    test_size: float
    random_state: int
    town: str
    remaining_lease: str
    with_outliers: List[str]
    with_rarelabels: List[str]
    skewed_features: List[str]
    by_onehot: List[str]
    by_ordinal: List[str]
    town_mappings: Dict[str, str]


class Config(BaseModel):
    """
    Master config object
    """

    app_config: AppConfig
    model_config: ModelConfig


def find_config_file() -> Path:
    """
    Locate the configuration file
    """
    if CONFIG_FILE_PATH.is_file():
        return CONFIG_FILE_PATH
    raise Exception(f"Config not found at {CONFIG_FILE_PATH}")


def fetch_config_from_yaml(cfg_path: Path = None) -> YAML:
    """
    Parse YAML containing the package configuration
    """
    if not cfg_path:
        cfg_path = find_config_file()

    if cfg_path:
        with open(cfg_path, "r") as conf_file:
            parsed_config = load(conf_file.read())
            return parsed_config
    raise OSError(f"Did not find config file at path: {cfg_path}")


def create_and_validate_config(parsed_config: YAML = None) -> Config:
    """
    Run validation on config values
    """
    if parsed_config is None:
        parsed_config = fetch_config_from_yaml()

    _config = Config(
        app_config=AppConfig(**parsed_config.data),
        model_config=ModelConfig(**parsed_config.data),
    )

    return _config


config = create_and_validate_config()



from feature_engine.encoding import OneHotEncoder, OrdinalEncoder, RareLabelEncoder
from feature_engine.outliers import Winsorizer
from feature_engine.transformation import YeoJohnsonTransformer
from sklearn.pipeline import Pipeline
from xgboost import XGBRegressor

from regression_model.config.core import config
from regression_model.processing import features as pp

housing_price_pipe = Pipeline(
    [
        # ====== PREPROCESSING ======
        # extract the number of years and months from the
        # variable ('remaining_lease')
        (
            "extract_numerical_values",
            pp.ExtractNumericalValue(variable=config.model_config.remaining_lease),
        ),
        # categorize the values of the variable ('towns')
        (
            "group_towns",
            pp.CombineTowns(
                variable=config.model_config.town,
                mappings=config.model_config.town_mappings,
            ),
        ),
        # ====== Numerical Encoding ======
        # remove outliers
        (
            "remove_outliers",
            Winsorizer(
                capping_method="iqr",
                tail="both",
                fold=1.5,
                variables=config.model_config.with_outliers,
            ),
        ),
        # perform Yeo-Johnson transformation on skewed features
        (
            "yeo_johnson_transformation",
            YeoJohnsonTransformer(variables=config.model_config.skewed_features),
        ),
        # ====== Categorical Encoding ======
        # rare labels encoding
        (
            "rare_label_encoding",
            RareLabelEncoder(
                n_categories=5, variables=config.model_config.with_rarelabels
            ),
        ),
        # one hot encoding
        (
            "one_hot_encoding",
            OneHotEncoder(variables=config.model_config.by_onehot, drop_last=True),
        ),
        # ordinal encoding
        ("ordinal_encoding", OrdinalEncoder(variables=config.model_config.by_ordinal)),
        # ====== Apply Model ======
        (
            "xgboost_regressor",
            XGBRegressor(random_state=config.model_config.random_state),
        ),
    ]
)



from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pydantic import BaseModel, ValidationError

from regression_model.config.core import config


def drop_na_inputs(*, input_data: pd.DataFrame) -> pd.DataFrame:
    """
    Check model inputs for na values and filter
    """
    validated_data = input_data.copy()
    new_vars_with_na = [
        var
        for var in config.model_config.features
        if validated_data[var].isnull().sum() > 0
    ]
    validated_data.dropna(subset=new_vars_with_na, inplace=True)

    return validated_data


def validate_inputs(*, input_data: pd.DataFrame) -> Tuple[pd.DataFrame, Optional[dict]]:
    """
    Check model inputs for unprocessable values
    """
    target_ = config.model_config.target
    if target_ in input_data.columns:
        input_data.loc[:, target_] = input_data[target_].apply(np.log)
    relevant_data = input_data[config.model_config.features].copy()
    validated_data = drop_na_inputs(input_data=relevant_data)
    errors = None

    try:
        MultipleHouseDataInputs(
            inputs=validated_data.replace({np.nan: None}).to_dict(orient="records")
        )
    except ValidationError as error:
        errors = error.json()

    return validated_data, errors


class HouseDataInputSchema(BaseModel):
    town: Optional[str]
    flat_type: Optional[str]
    storey_range: Optional[str]
    floor_area_sqm: Optional[float]
    flat_model: Optional[str]
    lease_commence_date: Optional[int]
    remaining_lease: Optional[str]
    dist_school: Optional[float]
    dist_mrt: Optional[float]
    dist_supermarket: Optional[float]
    dist_hawker: Optional[float]
    dist_npc: Optional[float]
    dist_central: Optional[float]
    id: Optional[int]


class MultipleHouseDataInputs(BaseModel):
    inputs: List[HouseDataInputSchema]



import typing as t
from pathlib import Path

import joblib
import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline

from regression_model import __version__ as _version
from regression_model.config.core import DATASET_DIR, TRAINED_MODEL_DIR, config


def load_dataset(*, file_name: str) -> pd.DataFrame:
    df = pd.read_csv(Path(f"{DATASET_DIR}/{file_name}"))

    target_ = config.model_config.target
    if target_ in df.columns:
        df.loc[:, target_] = df[target_].apply(np.log)

    df.drop(columns=config.model_config.dropped_feature, inplace=True)

    return df


def save_pipeline(*, pipeline_to_persist: Pipeline) -> None:
    """
    Persist the pipeline. Save the versioned model, and overwrites
    any previous saved models. This ensures that when the package is
    published, there is only one trained model that can be called,
    and we know exactly how it was built.
    """

    # Prepare verisoned save file name
    save_file_name = f"{config.app_config.pipeline_save_file}{_version}.pkl"
    save_path = TRAINED_MODEL_DIR / save_file_name

    removed_old_pipelines(files_to_keep=[save_file_name])
    joblib.dump(pipeline_to_persist, save_path)


def load_pipeline(*, file_name: str) -> Pipeline:
    """
    Load a persist pipeline
    """
    file_path = TRAINED_MODEL_DIR / file_name
    trained_model = joblib.load(filename=file_path)

    return trained_model


def removed_old_pipelines(*, files_to_keep: t.List[str]) -> None:
    """
    removed old model pipelines. This is ensure there is a simple
    one-to-one mapping between the package version and model
    version to be imported and used by other applications.
    """

    do_not_delete = files_to_keep + ["__init__.py"]
    for model_file in TRAINED_MODEL_DIR.iterdir():
        if model_file.name not in do_not_delete:
            model_file.unlink()



import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class ExtractNumericalValue(BaseEstimator, TransformerMixin):
    """
    Extract the number of years and months of the variable(remaining_lease)
    """

    def __init__(self, variable: str):
        super().__init__()

        self.variable = variable

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        # We need this step to fit the sklearn pipeline
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        X.loc[:, self.variable] = X[self.variable].apply(get_remaining_lease)
        return X


class CombineTowns(BaseEstimator, TransformerMixin):
    """
    Categorize the values of the variable(towns)
    """

    def __init__(self, variable: str, mappings: dict):
        super().__init__()

        self.variable = variable
        self.mappings = mappings

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.copy()
        X.loc[:, self.variable] = X[self.variable].map(self.mappings)
        return X


def get_remaining_lease(x) -> float:

    x = x.split()

    if len(x) == 2:
        return float(x[0])

    return round(float(x[0]) + float(x[2]) / 12, 3)
